import os
import sys
from searcHPV.general import *
from searcHPV.generate_call_fusion import *
import string

def genomeFusion(window,out_dir,virRef):

    bam = f'{out_dir}/alignment/alignment.RG.indelre.mkdup.sort.bam'
    check_file(bam)
    #make dir for output
    out_dir = f'{out_dir}/call_fusion'
    mkdir(out_dir)

    #find the virus_chrm list
    virus_chrm_list = []
    with open(virRef) as virRefFile:
        for each in virRefFile.read().rstrip().split('\n'):
            if '>' in each:
                virus_chrm_list.append(each.split()[0].replace('>',''))

    #identify fusion points on genome
    for virus_chrm in virus_chrm_list:
        res = define_fusion(bam,virus_chrm,out_dir) #generate out_dir/genome_fusion.txt

        #filter and cluster fusion points
        ##sort result
        #if there are invalid characters in the chromosome name:
        invalidCharacter = str(string.punctuation)
        virus_chrm_file_name = virus_chrm.replace(invalidCharacter,"_")
        os.system(f'(head -n 1 {res} && tail -n +2 {res} | sort -k3,3rn) > {out_dir}/{virus_chrm_file_name}.genome_fusion.sort.txt')


        #change format for cluster
        chrm_li = range(1,22)
        chrm_li = list(map(str,chrm_li))
        chrm_li+=['X','Y']

        with open(f'{out_dir}/{virus_chrm_file_name}.all.result','w') as output:   
                fusion_li = []
                with open(f'{out_dir}/{virus_chrm_file_name}.genome_fusion.sort.txt') as res:
                    res.readline()
                    for line in res.read().rstrip().split('\n'):
                        elements = line.rstrip().split('\t')
                        if elements != ['']:
                            #split read count
                            chrom = elements[0]
                            pos = elements[1]
                            count = elements[2]
                            #pair-end read count
                            pair_count = elements[3]
                            if chrom in chrm_li:
                                fusion_li.append(f'{elements[0]}:{elements[1]}:{count}:{pair_count}')
                to_print = ';'.join(fusion_li)
                output.write(f'{to_print}')

        ##cluster the events within 100bp from each other, maybe becasue of SVs or CNVs
        cluster_result(f'{out_dir}/{virus_chrm_file_name}.all.result',f'{out_dir}/{virus_chrm_file_name}.all.clustered.result',window)


        ##filter for sites with at least 2 split read  and 2 pairs of read support(high cutoff) and their summation greater than 5
        with open(f'{out_dir}/{virus_chrm_file_name}.all.clustered.result') as inf:
            with open(f'{out_dir}/{virus_chrm_file_name}.all.filtered.clustered.result','w') as outf:
                for line in inf.read().rstrip().split('\n'):
                    elements = line.rstrip()
                    if elements == "":
                        outf.write(elements)
                    else:
                        pos_li = elements.split(';')
                        new_pos_li = []
                        for pos in pos_li:
                            single_count = int(pos.split(':')[2])
                            pair_count = int(pos.split(':')[3])
                            if single_count > 2 and pair_count >2:
                            #or (single_count > 5) or (pair_count > 5)
                                new_pos_li.append(pos + ':high')
                            elif single_count + pair_count >= 5:
                                new_pos_li.append(pos + ':low')
                                
                        new_pos_infor = ';'.join(new_pos_li)
                        outf.write(new_pos_infor)
    return None


import sys
import os
import re
#config Dictionary formate:key1 = "Input File",key2 = "reference_file = ",value = full_path

#####################
##catenate reference genome
#Function: cat human genome + virus genome
#Parameters: humRef: human genome
#virRef: virus genome
#outputDir: output directory
def catRef(humRef, virRef, outputDir):
    os.system(f'cat {humRef} {virRef} > {outputDir}/hg_hpv.fa')
    newRef = f'{outputDir}/hg_hpv.fa'
    return(newRef)

#####################
##generate bash script for index reference genome
#Function: index customized reference
#bash_file: output bash script
#humRef: human genome
#virRef: virus genome
#newRef: customize genome
#outputDir: output directory
def indexRef(bash_file,ref):
    with open(bash_file,'w') as output:
        output.write(f'''bwa index {ref}
samtools faidx {ref}
picard CreateSequenceDictionary R={ref} O={ref.replace('.fa','.dict')}
''')
    return None




#####################
##generate alignment bash file
#Function: generate alignment bash file for a sample
#Parameters: bash_file: output bash file
#ref: reference gene
#fq1: fastq1 file
#fq2: fastq2 file
#out_dir: outputPath
#gz: if fastq file is in gz format: default = True
def generate_alignment_bash(bash_file,ref,fq1,fq2,out_dir,gz = True):
    with open(bash_file,'w') as output:
        if gz:
            output.write(f'''
    bwa mem {ref} '<zcat {fq1}' '<zcat {fq2}' > {out_dir}/alignment.sam
    samtools view -bhS {out_dir}/alignment.sam > {out_dir}/alignment.bam
    samtools sort {out_dir}/alignment.bam -o {out_dir}/alignment.sort.bam
    rm {out_dir}/alignment.sam
    echo \'alignment done\'
    ''')
        else:
            output.write(f'''
    bwa mem {ref} '<cat {fq1}' '<cat {fq2}' > {out_dir}/alignment.sam
    samtools view -bhS {out_dir}/alignment.sam > {out_dir}/alignment.bam
    samtools sort {out_dir}/alignment.bam -o {out_dir}/alignment.sort.bam
    rm {out_dir}/alignment.sam
    echo \'alignment done\'
    ''')
    return None

 ################


 ################
 ##generate indel re alignment bash file
#Parameters: bash_file: output bash file
#ref: reference gene
#fq1: fastq1 file
#fq2: fastq2 file
#out_dir: outputPath
#sample: samle name
#samtools: full path of samtools
#bwa: full path of bwa
#picard: full path of picard
#gatk: full path of GATK
#java: full path of java
def generate_indel_alignment_bash(bash_File,ref,out_dir):
    with open(bash_File,'w') as output:
        output.write(f'''picard \
AddOrReplaceReadGroups \
I={out_dir}/alignment.bam \
O={out_dir}/alignment.RG.bam \
RGID=sample \
RGPL=illumina \
RGPU=NA \
RGSM=sample \
RGLB=sample
samtools sort {out_dir}/alignment.RG.bam -o {out_dir}/alignment.RG.sort.bam
samtools index {out_dir}/alignment.RG.sort.bam
GenomeAnalysisTK \
-T RealignerTargetCreator \
-R {ref} \
-I {out_dir}/alignment.RG.sort.bam \
-o {out_dir}/alignment.RG.intervals
GenomeAnalysisTK \
-T IndelRealigner \
-R {ref} \
-I {out_dir}/alignment.RG.sort.bam \
-targetIntervals {out_dir}/alignment.RG.intervals \
-o {out_dir}/alignment.RG.indelre.bam
samtools index {out_dir}/alignment.RG.indelre.bam
echo \'indel alignment done\'''')



def generate_mkdup_bash(bash_File,out_dir):
    with open(bash_File,'w') as output:
        output.write(f'''
samtools sort -n {out_dir}/alignment.RG.indelre.bam -o {out_dir}/alignment.RG.indelre.sortbyQ.bam
picard MarkDuplicates \
I={out_dir}/alignment.RG.indelre.sortbyQ.bam \
O={out_dir}/alignment.RG.indelre.mkdup.bam \
M={out_dir}/alignment.RG.indelre.mkdup.txt \
TAGGING_POLICY=All ASSUME_SORT_ORDER=queryname
samtools sort {out_dir}/alignment.RG.indelre.mkdup.bam -o {out_dir}/alignment.RG.indelre.mkdup.sort.bam
samtools index {out_dir}/alignment.RG.indelre.mkdup.sort.bam
echo \'indel alignment done\'''')


#delete intermediate bam file
def rm_inter_bam(bash_File,out_dir):
    
    with open(bash_File,'w') as output:
        output.write(f'''rm {out_dir}/alignment.RG.indelre.mkdup.bam
rm {out_dir}/alignment.RG.indelre.bam*
rm {out_dir}/alignment.RG.indelre.bai
rm {out_dir}/alignment.RG.sort.bam*
rm {out_dir}/alignment.RG.indelre.sortbyQ.bam 
rm {out_dir}/alignment.RG.bam
rm {out_dir}/alignment.bam
rm {out_dir}/alignment.sort.bam
''')



############################










import os
import sys
import re
from searcHPV.general import *

#############
#Function: generate bash script for mapping contigs to human+hpv genome
#out_dir:out_dir for searcHPV
#humRef: human reference genome
#return: path of bash script 
def mapToRef(out_dir):
    newRef = f'{out_dir}/hg_hpv.fa'
    sitesPath = f'{out_dir}/assemble/'
    listSites = os.listdir(sitesPath)
    outputPath = f'{out_dir}/call_fusion_virus/'
    mkdir(outputPath)
    #if there are sites
    if listSites != ['']:
        with open( f'{outputPath}/alignContigsToRef.sh','w') as bashFile:
            bashFile.write('#!/bin/bash\n')
            for site in listSites:
                if ".sh" not in site:
                    contigPath = f'{sitesPath}/{site}/pearOutput/'            
                    sitePath = f'{out_dir}/call_fusion_virus/{site}/'
                    mkdir(sitePath)
                    bashFile.write( f'''bwa mem -R \'@RG\\tID:hpv\\tSM:hpv\\tLB:hpv\\tPL:ILLUMINA\' -M -t 8 \
{newRef} {contigPath}/{site}.all.fa.cap.contigs > {sitePath}/{site}.contig.sam;
samtools view -bhS {sitePath}/{site}.contig.sam > {sitePath}/{site}.contig.bam;
samtools sort {sitePath}/{site}.contig.bam -o {sitePath}/{site}.contig.sort.bam;
samtools index {sitePath}/{site}.contig.sort.bam;
rm {sitePath}/{site}.contig.bam;\n''')
    return f'{outputPath}/alignContigsToRef.sh'




#############
#Function: generate bash script for mapping contigs to human genome
#out_dir:out_dir for searcHPV
#humRef: human reference genome
#return: path of bash script 
def mapToHgRef(out_dir,humRef):
    sitesPath = f'{out_dir}/assemble/'
    listSites = os.listdir(sitesPath)
    outputPath = f'{out_dir}/call_fusion_virus/'
    mkdir(outputPath)
    #if there are sites
    if listSites != ['']:
        with open( f'{outputPath}/alignContigsToGenome.sh','w') as bashFile:
            bashFile.write('#!/bin/bash\n')
            for site in listSites:
                if ".sh" not in site:
                    contigPath = f'{sitesPath}/{site}/pearOutput/'            
                    sitePath = f'{out_dir}/call_fusion_virus/{site}/'
                    mkdir(sitePath)
                    bashFile.write( f'''bwa mem -R \'@RG\\tID:hpv\\tSM:hpv\\tLB:hpv\\tPL:ILLUMINA\' -M -t 8 \
{humRef} {contigPath}/{site}.all.fa.cap.contigs > {sitePath}/{site}.contigToGenome.sam;
samtools view -bhS {sitePath}/{site}.contigToGenome.sam > {sitePath}/{site}.contigToGenome.bam;
samtools sort {sitePath}/{site}.contigToGenome.bam -o {sitePath}/{site}.contigToGenome.sort.bam;
samtools index {sitePath}/{site}.contigToGenome.sort.bam;
rm {sitePath}/{site}.contigToGenome.bam;\n''')
    return f'{outputPath}/alignContigsToGenome.sh'

#############
#Function: generate bash script for mapping contigs to hpv genome
#out_dir:out_dir for searcHPV
#virRef: virus reference genome
#return: path of bash script 
def mapToVirRef(out_dir,virRef):
    sitesPath = f'{out_dir}/assemble/'
    listSites = os.listdir(sitesPath)
    outputPath = f'{out_dir}/call_fusion_virus/'
    mkdir(outputPath)
    #if there are sites
    if listSites != ['']:
        with open( f'{outputPath}/alignContigsToVirus.sh','w') as bashFile:
            bashFile.write('#!/bin/bash\n')
            for site in listSites:
                if ".sh" not in site:
                    contigPath = f'{sitesPath}/{site}/pearOutput/'
                    sitePath = f'{out_dir}/call_fusion_virus/{site}/'
                    bashFile.write( f'''bwa mem -R \'@RG\\tID:hpv\\tSM:hpv\\tLB:hpv\\tPL:ILLUMINA\' -M -t 8 \
{virRef} {contigPath}/{site}.all.fa.cap.contigs > {contigPath}/{site}.contigToVirus.sam;
samtools view -bhS {contigPath}/{site}.contigToVirus.sam >  {contigPath}/{site}.contigToVirus.bam;
samtools sort {contigPath}/{site}.contigToVirus.bam -o {sitePath}/{site}.contigToVirus.sort.bam;
samtools index {sitePath}/{site}.contigToVirus.sort.bam;
samtools faidx {contigPath}/{site}.all.fa.cap.contigs;
rm {contigPath}/{site}.contigToVirus.bam;\n''')
    return f'{outputPath}/alignContigsToVirus.sh'

#############
#Function: count number of supprotive reads for each contig
#out_dir:out_dir for searcHPV
def count_sr(out_dir):
    dirName = f'{out_dir}/assemble/'
    sites = os.listdir(dirName)

    key = ""
    rowList = []
    for site in sites:
        if ".sh" not in site:
            sampleDic = {}
            cap3File = f'{dirName}/{site}/cap3Output/{site}.CAP3'
            with open(cap3File) as eachFile:
                for i in range(0, 5):
                    eachFile.readline()

                for row in eachFile:
                    if re.match('^\*',row):
                        
                        key = site+"."+\
                            row.rstrip().strip(' ').split(" ")[1]+row.rstrip().strip(' ').split(" ")[2]#contig number
                        rowList = []
                    else:
                        rowList.append(row)
                    sampleDic[key] = rowList
                    if re.match("^DETAILED",row):
                        break

        #         for items in sampleDic.items():
        #            print(items)

            with open(f'{out_dir}/call_fusion_virus/{site}/srNum.txt', "w") as sampleOutput:
                for key, value in sampleDic.items():
                    sampleOutput.write(key+"\t"+str(len(value)-1)+'\n')
    return None

















from searcHPV.generate_alignment import *
import subprocess
from searcHPV.general import *
#####################
#pipeline of alignment
#Function:
#1. customize reference genome
#2. generate alignment bash script
#3. run alignment script
#fq1: paired-end raw sequencing data, fastq1
#fq2: paired-end raw sequencing data, fastq2
#humRef: human reference genome
#outputDir: output directory
#multi: if fastq file is in gz format: default = True
#index: if True, index the reference files; if False, not index the references files
def alignment(fq1, fq2, humRef, virRef, outputDir, index, gz):
    #make output dir
    outputDir = os.path.abspath(outputDir)
    mkdir(outputDir)
    scriptDir = f'{outputDir}/alignment'
    mkdir(scriptDir)
    #catenate humRef and virRef
    ref = catRef(humRef,virRef, outputDir)

    #generate alignment bash

    alignmentFile = scriptDir + "/orignal.alignment.sh"
    indelFile = scriptDir + "/indel.alignment.sh"
    generate_alignment_bash(alignmentFile,ref,fq1,fq2,scriptDir,gz)
    generate_indel_alignment_bash(indelFile,ref,scriptDir)
    check_file(alignmentFile)
    check_file(indelFile)
    bashFile = scriptDir + f"/alignment.sh"

    #remove intermediate file
    rmInter = scriptDir + f"/rm_inter.sh"
    rm_inter_bam(rmInter,scriptDir)

    if index:
        #index humRef and virRef
        indexFileHum = scriptDir + "/index_hum.sh"
        indexRef(indexFileHum,humRef)
        indexFileVir = scriptDir + "/index_vir.sh"
        indexRef(indexFileVir,virRef)
        indexFile = scriptDir + "/index.sh"
        indexRef(indexFile,ref)
    else:
        indexFile = scriptDir + "/index.sh"
        indexRef(indexFile,ref)

    if index:
    ##generate mkdup alignment bash file
        mkdupFile = scriptDir + "/mkdup.alignment.sh"
        generate_mkdup_bash(mkdupFile,scriptDir)

        with open(bashFile,'w') as output:
            output.write(f'''#!/bin/bash
bash {indexFileHum};
bash {indexFileVir};
bash {indexFile};
bash {alignmentFile};
bash {indelFile};
bash {mkdupFile};
bash {rmInter};''')
    else:
        ##generate mkdup alignment bash file
        mkdupFile = scriptDir + "/mkdup.alignment.sh"
        generate_mkdup_bash(mkdupFile,scriptDir)

        with open(bashFile,'w') as output:
            output.write(f'''#!/bin/bash
bash {indexFile};
bash {alignmentFile};
bash {indelFile};
bash {mkdupFile}
bash {rmInter};''')


    #run scripts
    check_file(bashFile)
    os.system(f'chmod +x {bashFile}')
    subprocess.call(bashFile)
